@@ -1,5 +1,5 @@
 ### Copyright (C) 2002-2006 Stephen Kennedy <stevek@gnome.org>
-### Copyright (C) 2009-2010 Kai Willadsen <kai.willadsen@gmail.com>
+### Copyright (C) 2009-2011 Kai Willadsen <kai.willadsen@gmail.com>
 
 ### This program is free software; you can redistribute it and/or modify
 ### it under the terms of the GNU General Public License as published by
@@ -22,7 +22,6 @@ import paths
 from ui import gnomeglade
 import gtk
 import gtk.keysyms
-import math
 import misc
 import os
 from gettext import gettext as _
@@ -219,6 +218,12 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         tree.STATE_MISSING: "delete",
     }
 
+    state_actions = {
+        tree.STATE_NORMAL: ("normal", "ShowSame"),
+        tree.STATE_NEW: ("new", "ShowNew"),
+        tree.STATE_MODIFIED: ("modified", "ShowModified"),
+    }
+
     def __init__(self, prefs, num_panes):
         melddoc.MeldDoc.__init__(self, prefs)
         gnomeglade.Component.__init__(self, paths.ui_dir("dirdiff.ui"), "dirdiff")
@@ -229,15 +234,13 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
             ("DirCopyRight", gtk.STOCK_GO_FORWARD,   _("Copy _Right"),    "<Alt>Right", _("Copy to right"), self.on_button_copy_right_clicked),
             ("DirDelete",    gtk.STOCK_DELETE,        None,         "Delete", _("Delete selected"), self.on_button_delete_clicked),
             ("Hide",         gtk.STOCK_NO,           _("Hide"),     None, _("Hide selected"), self.on_filter_hide_current_clicked),
-
-            ("DirOpen",      gtk.STOCK_OPEN,          None,         None, _("Open selected"), self.on_button_open_clicked),
         )
 
         toggleactions = (
             ("IgnoreCase",   gtk.STOCK_ITALIC,  _("Case"),     None, _("Ignore case of entries"), self.on_button_ignore_case_toggled, False),
-            ("ShowSame",     gtk.STOCK_APPLY,   _("Same"),     None, _("Show identical"), self.on_filter_state_normal_toggled, True),
-            ("ShowNew",      gtk.STOCK_ADD,     _("New"),      None, _("Show new"), self.on_filter_state_new_toggled, True),
-            ("ShowModified", gtk.STOCK_REMOVE,  _("Modified"), None, _("Show modified"), self.on_filter_state_modified_toggled, True),
+            ("ShowSame",     gtk.STOCK_APPLY,   _("Same"),     None, _("Show identical"), self.on_filter_state_toggled, False),
+            ("ShowNew",      gtk.STOCK_ADD,     _("New"),      None, _("Show new"), self.on_filter_state_toggled, False),
+            ("ShowModified", gtk.STOCK_REMOVE,  _("Modified"), None, _("Show modified"), self.on_filter_state_toggled, False),
 
             ("CustomFilterMenu", None, _("Filters"), None, _("Set active filters"), self.on_custom_filter_menu_toggled, False),
         )
@@ -267,6 +270,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
             self.focus_in_events.append(handler_id)
             handler_id = treeview.connect("focus-out-event", self.on_treeview_focus_out_event)
             self.focus_out_events.append(handler_id)
+        self.prev_path, self.next_path = None, None
         self.on_treeview_focus_out_event(None, None)
         self.treeview_focussed = None
 
@@ -287,20 +291,21 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
             self.scrolledwindow[i].get_vadjustment().connect("value-changed", self._sync_vscroll )
             self.scrolledwindow[i].get_hadjustment().connect("value-changed", self._sync_hscroll )
         self.linediffs = [[], []]
-        self.state_filters = [
-            tree.STATE_NORMAL,
-            tree.STATE_MODIFIED,
-            tree.STATE_NEW,
-        ]
 
-    def _custom_popup_deactivated(self, popup):
-        self.filter_menu_button.set_active(False)
+        self.state_filters = []
+        for s in self.state_actions:
+            if self.state_actions[s][0] in self.prefs.dir_status_filters:
+                self.state_filters.append(s)
+                action_name = self.state_actions[s][1]
+                self.actiongroup.get_action(action_name).set_active(True)
 
     def on_custom_filter_menu_toggled(self, item):
         if item.get_active():
-            self.custom_popup.connect("deactivate", self._custom_popup_deactivated)
+            self.custom_popup.connect("deactivate",
+                                      lambda popup: item.set_active(False))
             self.custom_popup.popup(None, None, misc.position_menu_under_widget,
-                                    1, gtk.get_current_event_time(), self.filter_menu_button)
+                                    1, gtk.get_current_event_time(),
+                                    self.filter_menu_button)
 
     def _cleanup_filter_menu_button(self, ui):
         if self.popup_deactivate_id:
@@ -326,8 +331,6 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         self._create_filter_menu_button(ui)
         self.ui_manager = ui
 
-        # FIXME: Add real sensitivity handling
-        self.emit("next-diff-changed", True, True)
         if self.treeview_focussed:
             self.scheduler.add_task(self.treeview_focussed.grab_focus)
             self.scheduler.add_task(self.on_treeview_cursor_changed)
@@ -706,6 +709,19 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         pane = self._get_focused_pane()
         if pane is None:
             return
+
+        cursor_path, cursor_col = self.treeview[pane].get_cursor()
+        if not cursor_path:
+            self.emit("next-diff-changed", False, False)
+        else:
+            # TODO: Only recalculate when needed;
+            # not self.prev_path < cursor_path < self.next_path is a start,
+            # but fails when we move off a changed row.
+            prev_path, next_path = self.model._find_next_prev_diff(cursor_path)
+            self.prev_path, self.next_path = prev_path, next_path
+            have_next_diffs = (prev_path is not None, next_path is not None)
+            self.emit("next-diff-changed", *have_next_diffs)
+
         paths = self._get_selected_paths(pane)
         if len(paths) > 0:
             def rwx(mode):
@@ -755,17 +771,19 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
 
     def on_treeview_row_activated(self, view, path, column):
         pane = self.treeview.index(view)
-        allrows = self.model.value_paths(self.model.get_iter(path))
+        rows = self.model.value_paths(self.model.get_iter(path))
         # Click a file: compare; click a directory: expand; click a missing
         # entry: check the next neighbouring entry
         pane_ordering = ((0, 1, 2), (1, 2, 0), (2, 1, 0))
         for p in pane_ordering[pane]:
-            if p < self.num_panes and os.path.exists(allrows[p]):
+            if p < self.num_panes and rows[p] and os.path.exists(rows[p]):
                 pane = p
                 break
-        if os.path.isfile(allrows[pane]):
-            self.emit("create-diff", [r for r in allrows if os.path.isfile(r)])
-        elif os.path.isdir(allrows[pane]):
+        if not rows[pane]:
+            return
+        if os.path.isfile(rows[pane]):
+            self.emit("create-diff", [r for r in rows if os.path.isfile(r)])
+        elif os.path.isdir(rows[pane]):
             if view.row_expanded(path):
                 view.collapse_row(path)
             else:
@@ -810,7 +828,8 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         self.copy_selected(1)
     def on_button_delete_clicked(self, button):
         self.delete_selected()
-    def on_button_open_clicked(self, button):
+
+    def open_external(self):
         pane = self._get_focused_pane()
         if pane is not None:
             m = self.model
@@ -820,21 +839,18 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
     def on_button_ignore_case_toggled(self, button):
         self.refresh()
 
-    def _update_state_filter(self, state, active):
-        assert state in (tree.STATE_NEW, tree.STATE_MODIFIED, tree.STATE_NORMAL)
-        try:
-            self.state_filters.remove( state )
-        except ValueError:
-            pass
-        if active:
-            self.state_filters.append( state )
+    def on_filter_state_toggled(self, button):
+        active_action = lambda a: self.actiongroup.get_action(a).get_active()
+        active_filters = [a for a in self.state_actions if \
+                          active_action(self.state_actions[a][1])]
+
+        if set(active_filters) == set(self.state_filters):
+            return
+
+        state_strs = [self.state_actions[s][0] for s in active_filters]
+        self.state_filters = active_filters
+        self.prefs.dir_status_filters = state_strs
         self.refresh()
-    def on_filter_state_normal_toggled(self, button):
-        self._update_state_filter( tree.STATE_NORMAL, button.get_active() )
-    def on_filter_state_new_toggled(self, button):
-        self._update_state_filter( tree.STATE_NEW, button.get_active() )
-    def on_filter_state_modified_toggled(self, button):
-        self._update_state_filter( tree.STATE_MODIFIED, button.get_active() )
 
     def _update_name_filter(self, button, idx):
         self.name_filters[idx].active = button.get_active()
@@ -1072,12 +1088,12 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
             for component in changed:
                 child = model.iter_children( it )
                 while child:
-                    name = os.path.basename(model.value_path(child, pane))
-                    if component == name : # found it
+                    child_path = model.value_path(child, pane)
+                    # Found the changed path
+                    if child_path and component == os.path.basename(child_path):
                         it = child
                         break
-                    else:
-                        child = self.model.iter_next( child ) # next
+                    child = self.model.iter_next(child)
                 if not it:
                     break
             # save if found and unique
@@ -1091,20 +1107,16 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
 
     def next_diff(self, direction):
         if self.treeview_focussed:
-            pane = self.treeview.index( self.treeview_focussed )
+            pane = self.treeview.index(self.treeview_focussed)
         else:
             pane = 0
-        start_iter = self.model.get_iter( (self._get_selected_paths(pane) or [(0,)])[-1] )
-
-        search = {gtk.gdk.SCROLL_UP : self.model.inorder_search_up}.get(direction, self.model.inorder_search_down)
-        for it in search( start_iter ):
-            state = self.model.get_state(it, pane)
-            if state not in (tree.STATE_NORMAL, tree.STATE_EMPTY):
-                curpath = self.model.get_path(it)
-                self.treeview[pane].expand_to_path(curpath)
-                self.treeview[pane].set_cursor(curpath)
-                return
+        if direction == gtk.gdk.SCROLL_UP:
+            path = self.prev_path
+        else:
+            path = self.next_path
+        if path:
+            self.treeview[pane].expand_to_path(path)
+            self.treeview[pane].set_cursor(path)
 
     def on_reload_activate(self, *extra):
         self.on_fileentry_activate(None)
-
