@@ -19,13 +19,11 @@ import codecs
 import copy
 import os
 from gettext import gettext as _
-import re
-import difflib
-import struct
 import sys
 import time
 
 import pango
+import glib
 import gobject
 import gtk
 import gtk.keysyms
@@ -42,7 +40,6 @@ import merge
 
 from meldapp import app
 from util.sourceviewer import srcviewer
-from util.namedtuple import namedtuple
 
 
 class CachedSequenceMatcher(object):
@@ -129,7 +126,7 @@ class BufferLines(object):
                 # It's possible that the last line in a file would end in a
                 # line break character, which requires no joining.
                 if end and end[-1] in additional_breaks and \
-                   line and line[-1] not in additional_breaks:
+                   (not line or line[-1] not in additional_breaks):
                     assert len(ends) >= i + 1
                     lines[i:i + 2] = [line + end[-1] + lines[i + 1]]
                     ends[i:i + 2] = [end + ends[i + 1]]
@@ -161,17 +158,19 @@ MASK_SHIFT, MASK_CTRL = 1, 2
 
 MODE_REPLACE, MODE_DELETE, MODE_INSERT = 0, 1, 2
 
-def get_iter_at_line_or_eof(buffer, line):
-    if line >= buffer.get_line_count():
-        return buffer.get_end_iter()
-    return buffer.get_iter_at_line(line)
+def get_iter_at_line_or_eof(buf, line):
+    if line >= buf.get_line_count():
+        return buf.get_end_iter()
+    return buf.get_iter_at_line(line)
 
-def insert_with_tags_by_name(buffer, line, text, tag):
-    if line >= buffer.get_line_count():
+def buffer_insert(buf, line, text):
+    if line >= buf.get_line_count():
         # TODO: We need to insert a linebreak here, but there is no
         # way to be certain what kind of linebreak to use.
         text = "\n" + text
-    buffer.insert_with_tags_by_name(get_iter_at_line_or_eof(buffer, line), text, tag)
+    it = get_iter_at_line_or_eof(buf, line)
+    buf.insert(it, text)
+    return it
 
 class CursorDetails(object):
     __slots__ = ("pane", "pos", "line", "offset", "chunk", "prev", "next",
@@ -190,6 +189,19 @@ class TaskEntry(object):
             setattr(self, var, val)
 
 
+class TextviewLineAnimation(object):
+    __slots__ = ("start_mark", "end_mark", "start_rgba", "end_rgba",
+                 "start_time", "duration")
+
+    def __init__(self, mark0, mark1, rgba0, rgba1, duration):
+        self.start_mark = mark0
+        self.end_mark = mark1
+        self.start_rgba = rgba0
+        self.end_rgba = rgba1
+        self.start_time = glib.get_current_time()
+        self.duration = duration
+
+
 class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
     """Two or three way diff of text files.
     """
@@ -252,39 +264,33 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         self._scroll_lock = False
         self.linediffer = self.differ()
         self.linediffer.ignore_blanks = self.prefs.ignore_blank_lines
-        self.in_nested_action = False
         self.in_nested_textview_gutter_expose = False
         self._inline_cache = set()
         self._cached_match = CachedSequenceMatcher()
+        self.anim_source_id = []
+        self.animating_chunks = []
         for buf in self.textbuffer:
-            buf.create_tag("edited line",   background = self.prefs.color_edited_bg,
-                                            foreground = self.prefs.color_edited_fg)
-            buf.create_tag("delete line",   background = self.prefs.color_delete_bg,
-                                            foreground = self.prefs.color_delete_fg)
-            buf.create_tag("replace line",  background = self.prefs.color_replace_bg,
-                                            foreground = self.prefs.color_replace_fg)
-            buf.create_tag("conflict line", background = self.prefs.color_conflict_bg,
-                                            foreground = self.prefs.color_conflict_fg)
-            buf.create_tag("inline line",   background = self.prefs.color_inline_bg,
-                                            foreground = self.prefs.color_inline_fg)
+            buf.create_tag("inline", background=self.prefs.color_inline_bg,
+                                     foreground=self.prefs.color_inline_fg)
+            self.anim_source_id.append(None)
+            self.animating_chunks.append([])
 
         def parse_to_cairo(color_spec):
-            color = gtk.gdk.color_parse(color_spec)
-            return [x / 65535. for x in (color.red, color.green, color.blue)]
+            c = gtk.gdk.color_parse(color_spec)
+            return tuple([x / 65535. for x in (c.red, c.green, c.blue)])
 
         self.fill_colors = {"insert"   : parse_to_cairo(self.prefs.color_delete_bg),
                             "delete"   : parse_to_cairo(self.prefs.color_delete_bg),
                             "conflict" : parse_to_cairo(self.prefs.color_conflict_bg),
                             "replace"  : parse_to_cairo(self.prefs.color_replace_bg)}
 
-        darken = lambda color: [x * 0.8 for x in color]
+        darken = lambda color: tuple([x * 0.8 for x in color])
         self.line_colors = {"insert"   : darken(self.fill_colors["insert"]),
                             "delete"   : darken(self.fill_colors["delete"]),
                             "conflict" : darken(self.fill_colors["conflict"]),
                             "replace"  : darken(self.fill_colors["replace"])}
 
         actions = (
-            ("FileOpen",          gtk.STOCK_OPEN,       None,            None, _("Open selected"), self.on_open_activate),
             ("MakePatch", None, _("Format as patch..."), None, _("Create a patch using differences between files"), self.make_patch),
             ("PrevConflict", None, _("Previous conflict"), "<Ctrl>I", _("Go to the previous conflict"), lambda x: self.on_next_conflict(gtk.gdk.SCROLL_UP)),
             ("NextConflict", None, _("Next conflict"), "<Ctrl>K", _("Go to the next conflict"), lambda x: self.on_next_conflict(gtk.gdk.SCROLL_DOWN)),
@@ -386,6 +392,11 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
                               self.on_cursor_position_changed)
             buf.handlers = id0, id1, id2, id3, id4
 
+    # Abbreviations for insert and overwrite that fit in the status bar
+    _insert_overwrite_text = (_("INS"), _("OVR"))
+    # Abbreviation for line, column so that it will fit in the status bar
+    _line_column_text = _("Ln %i, Col %i")
+
     def on_cursor_position_changed(self, buf, pspec, force=False):
         pane = self.textbuffer.index(buf)
         pos = buf.props.cursor_position
@@ -397,10 +408,8 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         offset = cursor_it.get_line_offset()
         line = cursor_it.get_line()
 
-        # Abbreviations for insert and overwrite that fit in the status bar
-        insert_overwrite = (_("INS"), _("OVR"))[self.textview_overwrite]
-        # Abbreviation for line, column so that it will fit in the status bar
-        line_column = _("Ln %i, Col %i") % (line + 1, offset + 1)
+        insert_overwrite = self._insert_overwrite_text[self.textview_overwrite]
+        line_column = self._line_column_text % (line + 1, offset + 1)
         status = "%s : %s" % (insert_overwrite, line_column)
         self.emit("status-changed", status)
 
@@ -822,7 +831,7 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
             if response==gtk.RESPONSE_OK:
                 for i in range(self.num_panes):
                     if try_save[i]:
-                        if self.save_file(i) != melddoc.RESULT_OK:
+                        if not self.save_file(i):
                             return gtk.RESPONSE_CANCEL
             elif response == gtk.RESPONSE_DELETE_EVENT:
                 response = gtk.RESPONSE_CANCEL
@@ -832,12 +841,10 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         # text buffer undo/redo
         #
     def on_textbuffer__begin_user_action(self, *buffer):
-        self.in_nested_action = True
         self.undosequence.begin_group()
 
     def on_textbuffer__end_user_action(self, *buffer):
         self.undosequence.end_group()
-        self.in_nested_action = False
         self.update_highlighting()
 
     def on_text_insert_text(self, buf, it, text, textlen):
@@ -860,7 +867,7 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         #
         #
 
-    def on_open_activate(self, *args):
+    def open_external(self):
         pane = self._get_focused_pane()
         if pane >= 0:
             if self.bufferdata[pane].filename:
@@ -1085,11 +1092,23 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         self.update_highlighting()
         self._connect_buffer_handlers()
         self._set_merge_action_sensitivity()
+
+        langs = []
         for i in range(self.num_panes):
             filename = self.bufferdata[i].filename
             if filename:
-                lang = srcviewer.get_language_from_file(filename)
-                srcviewer.set_language(self.textbuffer[i], lang)
+                langs.append(srcviewer.get_language_from_file(filename))
+            else:
+                langs.append(None)
+
+        # If we have only one identified language then we assume that all of
+        # the files are actually of that type.
+        real_langs = [l for l in langs if l]
+        if real_langs and real_langs.count(real_langs[0]) == len(real_langs):
+            langs = (real_langs[0],) * len(langs)
+
+        for i in range(self.num_panes):
+            srcviewer.set_language(self.textbuffer[i], langs[i])
             srcviewer.set_highlight_syntax(self.textbuffer[i],
                                            self.prefs.use_syntax_highlighting)
         yield 0
@@ -1124,7 +1143,8 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
                 # are active, and may be altering the comparison. It would be
                 # better if we only showed this message if the filters *did*
                 # change the text in question.
-                if self.text_filters:
+                active_filters = any([f.active for f in self.text_filters])
+                if active_filters:
                     secondary_text = _("Text filters are being used, and may "
                                        "be masking differences between files. "
                                        "Would you like to compare the "
@@ -1140,7 +1160,7 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
                 if index == 0:
                     button.props.label = _("Hi_de")
 
-                if self.text_filters:
+                if active_filters:
                     msgarea.add_button(_("Show without filters"),
                                        gtk.RESPONSE_OK)
 
@@ -1160,12 +1180,12 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
             self.set_files([None] * self.num_panes)
 
     def update_highlighting(self):
-        if not self.in_nested_action:
+        if not self.undosequence.in_grouped_action():
             self.scheduler.add_task(self._update_highlighting().next)
 
     def _update_highlighting(self):
         alltexts = self.buffer_texts
-        alltags = [b.get_tag_table().lookup("inline line") for b in self.textbuffer]
+        alltags = [b.get_tag_table().lookup("inline") for b in self.textbuffer]
         progress = [b.create_mark("progress", b.get_start_iter()) for b in self.textbuffer]
         newcache = set()
         for chunk in self.linediffer.all_changes():
@@ -1246,8 +1266,8 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         area = event.area
         x, y = textview.window_to_buffer_coords(gtk.TEXT_WINDOW_WIDGET,
                                                 area.x, area.y)
-        bounds = (self._pixel_to_line(pane, y),
-                  self._pixel_to_line(pane, y + area.height + 1))
+        bounds = (textview.get_line_num_for_y(y),
+                  textview.get_line_num_for_y(y + area.height + 1))
 
         width, height = textview.allocation.width, textview.allocation.height
         context = event.window.cairo_create()
@@ -1256,8 +1276,8 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         context.set_line_width(1.0)
 
         for change in self.linediffer.single_changes(pane, bounds):
-            ypos0 = self._line_to_pixel(pane, change[1]) - visible.y
-            ypos1 = self._line_to_pixel(pane, change[2]) - visible.y
+            ypos0 = textview.get_y_for_line_num(change[1]) - visible.y
+            ypos1 = textview.get_y_for_line_num(change[2]) - visible.y
 
             context.rectangle(-0.5, ypos0 - 0.5, width + 1, ypos1 - ypos0)
             if change[1] != change[2]:
@@ -1277,6 +1297,42 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
             context.rectangle(0, ypos - visible.y, width, line_height)
             context.fill()
 
+        current_time = glib.get_current_time()
+        new_anim_chunks = []
+        for c in self.animating_chunks[pane]:
+            percent = min(1.0, (current_time - c.start_time) / c.duration)
+            rgba_pairs = zip(c.start_rgba, c.end_rgba)
+            rgba = [s + (e - s) * percent for s, e in rgba_pairs]
+
+            it = self.textbuffer[pane].get_iter_at_mark(c.start_mark)
+            ystart, _ = self.textview[pane].get_line_yrange(it)
+            it = self.textbuffer[pane].get_iter_at_mark(c.end_mark)
+            yend, _ = self.textview[pane].get_line_yrange(it)
+            if ystart == yend:
+                ystart -= 1
+
+            context.set_source_rgba(*rgba)
+            context.rectangle(0, ystart - visible.y, width, yend - ystart)
+            context.fill()
+
+            if current_time <= c.start_time + c.duration:
+                new_anim_chunks.append(c)
+            else:
+                self.textbuffer[pane].delete_mark(c.start_mark)
+                self.textbuffer[pane].delete_mark(c.end_mark)
+        self.animating_chunks[pane] = new_anim_chunks
+
+        if self.animating_chunks[pane] and self.anim_source_id[pane] is None:
+            def anim_cb():
+                textview.queue_draw()
+                return True
+            # Using timeout_add interferes with recalculation of inline
+            # highlighting; this mechanism could be improved.
+            self.anim_source_id[pane] = gobject.idle_add(anim_cb)
+        elif not self.animating_chunks[pane] and self.anim_source_id[pane]:
+            gobject.source_remove(self.anim_source_id[pane])
+            self.anim_source_id[pane] = None
+
         if event.window == textview.get_window(gtk.TEXT_WINDOW_LEFT):
             self.in_nested_textview_gutter_expose = True
             textview.emit("expose-event", event)
@@ -1323,7 +1379,7 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
                 self.fileentry[pane].set_filename( bufdata.filename)
                 self.fileentry[pane].prepend_history(bufdata.filename)
             else:
-                return melddoc.RESULT_ERROR
+                return False
         start, end = buf.get_bounds()
         text = unicode(buf.get_text(start, end, False), 'utf8')
         if bufdata.newlines:
@@ -1350,15 +1406,15 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
                 if misc.run_dialog(
                     _("'%s' contains characters not encodable with '%s'\nWould you like to save as UTF-8?") % (bufdata.label, bufdata.encoding),
                     self, gtk.MESSAGE_ERROR, gtk.BUTTONS_YES_NO) != gtk.RESPONSE_YES:
-                    return melddoc.RESULT_ERROR
+                    return False
 
         save_to = bufdata.savefile or bufdata.filename
         if self._save_text_to_filename(save_to, text):
             self.emit("file-changed", save_to)
             self.undosequence.checkpoint(buf)
-            return melddoc.RESULT_OK
+            return True
         else:
-            return melddoc.RESULT_ERROR
+            return False
 
     def make_patch(self, *extra):
         dialog = patchdialog.PatchDialog(self)
@@ -1561,26 +1617,6 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
             self.queue_draw()
             self.recompute_label()
 
-    def _line_to_pixel(self, pane, line ):
-        it = self.textbuffer[pane].get_iter_at_line(line)
-        y, h = self.textview[pane].get_line_yrange(it)
-        if line >= self.textbuffer[pane].get_line_count():
-            return y + h - 1
-        return y
-
-    def _pixel_to_line(self, pane, pixel ):
-        return self.textview[pane].get_line_at_y( pixel )[0].get_line()
-
-    def _find_next_chunk(self, direction, pane):
-        if direction == gtk.gdk.SCROLL_DOWN:
-            target = self.cursor.next
-        else: # direction == gtk.gdk.SCROLL_UP
-            target = self.cursor.prev
-
-        if target is None:
-            return None
-        return self.linediffer.get_chunk(target, pane)
-
     def next_diff(self, direction):
         pane = self._get_focused_pane()
         if pane == -1:
@@ -1590,7 +1626,15 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
                 pane = 0
         buf = self.textbuffer[pane]
 
-        c = self._find_next_chunk(direction, pane)
+        if direction == gtk.gdk.SCROLL_DOWN:
+            target = self.cursor.next
+        else: # direction == gtk.gdk.SCROLL_UP
+            target = self.cursor.prev
+
+        if target is None:
+            return
+
+        c = self.linediffer.get_chunk(target, pane)
         if c:
             # Warp the cursor to the first line of next chunk
             if self.cursor.line != c[1]:
@@ -1602,15 +1646,28 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         start = get_iter_at_line_or_eof(b0, chunk[1])
         end = get_iter_at_line_or_eof(b0, chunk[2])
         t0 = unicode(b0.get_text(start, end, False), 'utf8')
+
         if copy_up:
             if chunk[2] >= b0.get_line_count() and \
                chunk[3] < b1.get_line_count():
                 # TODO: We need to insert a linebreak here, but there is no
                 # way to be certain what kind of linebreak to use.
                 t0 = t0 + "\n"
-            insert_with_tags_by_name(b1, chunk[3], t0, "edited line")
+            dst_start = get_iter_at_line_or_eof(b1, chunk[3])
+            mark0 = b1.create_mark(None, dst_start, True)
+            new_end = buffer_insert(b1, chunk[3], t0)
         else: # copy down
-            insert_with_tags_by_name(b1, chunk[4], t0, "edited line")
+            dst_start = get_iter_at_line_or_eof(b1, chunk[4])
+            mark0 = b1.create_mark(None, dst_start, True)
+            new_end = buffer_insert(b1, chunk[4], t0)
+
+        mark1 = b1.create_mark(None, new_end, True)
+        # FIXME: If the inserted chunk ends up being an insert chunk, then
+        # this animation is not visible; this happens often in three-way diffs
+        rgba0 = self.fill_colors['insert'] + (1.0,)
+        rgba1 = self.fill_colors['insert'] + (0.0,)
+        anim = TextviewLineAnimation(mark0, mark1, rgba0, rgba1, 0.5)
+        self.animating_chunks[dst].append(anim)
 
     def replace_chunk(self, src, dst, chunk):
         b0, b1 = self.textbuffer[src], self.textbuffer[dst]
@@ -1619,10 +1676,18 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         dst_start = get_iter_at_line_or_eof(b1, chunk[3])
         dst_end = get_iter_at_line_or_eof(b1, chunk[4])
         t0 = unicode(b0.get_text(src_start, src_end, False), 'utf8')
+        mark0 = b1.create_mark(None, dst_start, True)
         self.on_textbuffer__begin_user_action()
         b1.delete(dst_start, dst_end)
-        insert_with_tags_by_name(b1, chunk[3], t0, "edited line")
+        new_end = buffer_insert(b1, chunk[3], t0)
         self.on_textbuffer__end_user_action()
+        mark1 = b1.create_mark(None, new_end, True)
+        # FIXME: If the inserted chunk ends up being an insert chunk, then
+        # this animation is not visible; this happens often in three-way diffs
+        rgba0 = self.fill_colors['insert'] + (1.0,)
+        rgba1 = self.fill_colors['insert'] + (0.0,)
+        anim = TextviewLineAnimation(mark0, mark1, rgba0, rgba1, 0.5)
+        self.animating_chunks[dst].append(anim)
 
     def delete_chunk(self, src, chunk):
         b0 = self.textbuffer[src]
@@ -1630,6 +1695,14 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         if chunk[2] >= b0.get_line_count():
             it.backward_char()
         b0.delete(it, get_iter_at_line_or_eof(b0, chunk[2]))
+        mark0 = b0.create_mark(None, it, True)
+        mark1 = b0.create_mark(None, it, True)
+        # TODO: Need a more specific colour here; conflict is wrong
+        rgba0 = self.fill_colors['conflict'] + (1.0,)
+        rgba1 = self.fill_colors['conflict'] + (0.0,)
+        anim = TextviewLineAnimation(mark0, mark1, rgba0, rgba1, 0.5)
+        self.animating_chunks[src].append(anim)
+
 
 ################################################################################
 #
