@@ -1,4 +1,5 @@
 ### Copyright (C) 2002-2009 Stephen Kennedy <stevek@gnome.org>
+### Copyright (C) 2010-2011 Kai Willadsen <kai.willadsen@gmail.com>
 
 ### This program is free software; you can redistribute it and/or modify
 ### it under the terms of the GNU General Public License as published by
@@ -20,6 +21,8 @@ from gettext import gettext as _
 import gtk
 
 from ui import gnomeglade
+from ui import listwidget
+import meldapp
 import misc
 import paths
 from util import prefs
@@ -28,100 +31,67 @@ import vc
 from util.sourceviewer import srcviewer
 
 
-class ListWidget(gnomeglade.Component):
-    def __init__(self, columns, prefs, key):
-        gnomeglade.Component.__init__(self, paths.ui_dir("preferences.glade"), "listwidget")
+class FilterList(listwidget.ListWidget):
+
+    def __init__(self, prefs, key, filter_type):
+        default_entry = [_("label"), False, _("pattern"), True]
+        listwidget.ListWidget.__init__(self, default_entry)
         self.prefs = prefs
         self.key = key
-        self.treeview.set_model( gtk.ListStore( *[c[1] for c in columns] ) )
-        view = self.treeview
-        def addTextCol(label, colnum):
-            model = view.get_model()
-            rentext = gtk.CellRendererText()
-            rentext.props.editable = 1
-            def change_text(ren, path, text):
-                model[path][colnum] = text
-                self._update_filter_string()
-            rentext.connect("edited", change_text)
-            column = gtk.TreeViewColumn(label, rentext, text=colnum)
-            view.append_column(column)
-        def addToggleCol(label, colnum):
-            model = view.get_model()
-            rentoggle = gtk.CellRendererToggle()
-            def change_toggle(ren, path):
-                model[path][colnum] = not ren.get_active()
-                self._update_filter_string()
-            rentoggle.connect("toggled", change_toggle)
-            column = gtk.TreeViewColumn(label, rentoggle, active=colnum)
-            view.append_column(column)
-        for c,i in zip( columns, range(len(columns))):
-            if c[1] == type(""):
-                addTextCol(c[0], i)
-            elif c[1] == type(0):
-                addToggleCol( c[0], 1)
-        view.get_selection().connect('changed', self._update_sensitivity)
-        view.get_model().connect('row-inserted', self._update_sensitivity)
-        view.get_model().connect('rows-reordered', self._update_sensitivity)
+        self.filter_type = filter_type
+
+        self.pattern_column.set_cell_data_func(self.validity_renderer,
+                                               self.valid_icon_celldata)
+
+        for filtstring in getattr(self.prefs, self.key).split("\n"):
+            filt = meldapp.FilterEntry.parse(filtstring, filter_type)
+            if filt is None:
+                continue
+            valid = filt.filter is not None
+            self.model.append([filt.label, filt.active,
+                               filt.filter_string, valid])
+
+        for signal in ('row-changed', 'row-deleted', 'row-inserted',
+                       'rows-reordered'):
+            self.model.connect(signal, self._update_filter_string)
+
         self._update_sensitivity()
-        self._update_filter_model()
 
-    def _update_sensitivity(self, *args):
-        (model, it, path) = self._get_selected()
-        if not it:
-            self.item_delete.set_sensitive(False)
-            self.item_up.set_sensitive(False)
-            self.item_down.set_sensitive(False)
-        else:
-            self.item_delete.set_sensitive(True)
-            self.item_up.set_sensitive(path > 0)
-            self.item_down.set_sensitive(path < len(model) - 1)
+    def valid_icon_celldata(self, col, cell, model, it, user_data=None):
+        is_valid = model.get_value(it, 3)
+        icon_name = "gtk-dialog-warning" if not is_valid else None
+        cell.set_property("stock-id", icon_name)
 
-    def on_item_new_clicked(self, button):
-        model = self.treeview.get_model()
-        model.append([_("label"), 0, _("pattern")])
-        self._update_filter_string()
-    def _get_selected(self):
-        (model, it) = self.treeview.get_selection().get_selected()
-        if it:
-            path = model.get_path(it)[0]
-        else:
-            path = None
-        return (model, it, path)
-    def on_item_delete_clicked(self, button):
-        (model, it, path) = self._get_selected()
-        model.remove(it)
-        self._update_filter_string()
-    def on_item_up_clicked(self, button):
-        (model, it, path) = self._get_selected()
-        model.swap(it, model.get_iter(path - 1))
-        self._update_filter_string()
-    def on_item_down_clicked(self, button):
-        (model, it, path) = self._get_selected()
-        model.swap(it, model.get_iter(path + 1))
-        self._update_filter_string()
-    def on_items_revert_clicked(self, button):
-        setattr( self.prefs, self.key, self.prefs.get_default(self.key) )
-        self._update_filter_model()
-    def _update_filter_string(self):
-        model = self.treeview.get_model()
+    def on_name_edited(self, ren, path, text):
+        self.model[path][0] = text
+
+    def on_cellrenderertoggle_toggled(self, ren, path):
+        self.model[path][1] = not ren.get_active()
+
+    def on_pattern_edited(self, ren, path, text):
+        filt = meldapp.FilterEntry.compile_filter(text, self.filter_type)
+        valid = filt is not None
+        self.model[path][2] = text
+        self.model[path][3] = valid
+
+    def _update_filter_string(self, *args):
         pref = []
-        for row in model:
-            pref.append("%s\t%s\t%s" % (row[0], row[1], row[2]))
-        setattr( self.prefs, self.key, "\n".join(pref) )
-    def _update_filter_model(self):
-        model = self.treeview.get_model()
-        model.clear()
-        for filtstring in getattr( self.prefs, self.key).split("\n"):
-            filt = misc.ListItem(filtstring)
-            model.append([filt.name, filt.active, filt.value])
-   
+        for row in self.model:
+            pattern = row[2]
+            if pattern:
+                pattern = pattern.replace('\r', '')
+                pattern = pattern.replace('\n', '')
+            pref.append("%s\t%s\t%s" % (row[0], 1 if row[1] else 0, pattern))
+        setattr(self.prefs, self.key, "\n".join(pref))
+
 
 class PreferencesDialog(gnomeglade.Component):
 
-    def __init__(self, parentapp):
-        gnomeglade.Component.__init__(self, paths.ui_dir("preferences.glade"), "preferencesdialog")
-        self.widget.set_transient_for(parentapp.widget)
-        self.prefs = parentapp.prefs
+    def __init__(self, parent, prefs):
+        gnomeglade.Component.__init__(self, paths.ui_dir("preferences.ui"),
+                                      "preferencesdialog", ["adjustment1"])
+        self.widget.set_transient_for(parent)
+        self.prefs = prefs
         if not self.prefs.use_custom_font:
             self.checkbutton_default_font.set_active(True)
             self.fontpicker.set_sensitive(False)
@@ -134,16 +104,17 @@ class PreferencesDialog(gnomeglade.Component):
         if srcviewer.gsv is not None:
             self.checkbutton_spaces_instead_of_tabs.set_active( self.prefs.spaces_instead_of_tabs )
             self.checkbutton_show_line_numbers.set_active( self.prefs.show_line_numbers )
+            self.checkbutton_show_whitespace.set_active(self.prefs.show_whitespace)
             self.checkbutton_use_syntax_highlighting.set_active( self.prefs.use_syntax_highlighting )
         else:
-            self.checkbutton_spaces_instead_of_tabs.set_sensitive(False)
-            self.checkbutton_show_line_numbers.set_sensitive(False)
-            self.checkbutton_use_syntax_highlighting.set_sensitive(False)
-            if gtk.pygtk_version >= (2, 12, 0):
-                no_sourceview_text = _("Only available if you have gnome-python-desktop installed")
-                self.checkbutton_spaces_instead_of_tabs.set_tooltip_text(no_sourceview_text)
-                self.checkbutton_show_line_numbers.set_tooltip_text(no_sourceview_text)
-                self.checkbutton_use_syntax_highlighting.set_tooltip_text(no_sourceview_text)
+            no_sourceview_text = \
+                _("Only available if you have gnome-python-desktop installed")
+            for w in (self.checkbutton_spaces_instead_of_tabs,
+                      self.checkbutton_show_line_numbers,
+                      self.checkbutton_use_syntax_highlighting,
+                      self.checkbutton_show_whitespace):
+                w.set_sensitive(False)
+                w.set_tooltip_text(no_sourceview_text)
         # TODO: This doesn't restore the state of character wrapping when word
         # wrapping is disabled, but this is hard with our existing gconf keys
         if self.prefs.edit_wrap_lines != gtk.WRAP_NONE:
@@ -159,23 +130,24 @@ class PreferencesDialog(gnomeglade.Component):
                       self.prefs.edit_command_type == "gnome"
         self.system_editor_checkbutton.set_active(use_default)
         self.custom_edit_command_entry.set_sensitive(not use_default)
-        self.custom_edit_command_entry.set_text( " ".join(self.prefs.get_custom_editor_command([])) )
+        custom_command = " ".join(self.prefs.get_editor_command([], "custom"))
+        self.custom_edit_command_entry.set_text(custom_command)
 
         # file filters
-        cols = [ (_("Name"), type("")), (_("Active"), type(0)), (_("Pattern"), type("")) ]
-        self.filefilter = ListWidget( cols, self.prefs, "filters")
+        self.filefilter = FilterList(self.prefs, "filters",
+                                     meldapp.FilterEntry.SHELL)
         self.file_filters_tab.pack_start(self.filefilter.widget)
         self.checkbutton_ignore_symlinks.set_active( self.prefs.ignore_symlinks)
+
         # text filters
-        cols = [ (_("Name"), type("")), (_("Active"), type(0)), (_("Regex"), type("")) ]
-        self.textfilter = ListWidget( cols, self.prefs, "regexes")
+        self.textfilter = FilterList(self.prefs, "regexes",
+                                     meldapp.FilterEntry.REGEX)
         self.text_filters_tab.pack_start(self.textfilter.widget)
         self.checkbutton_ignore_blank_lines.set_active( self.prefs.ignore_blank_lines )
         # encoding
         self.entry_text_codecs.set_text( self.prefs.text_codecs )
-    #
-    # editor
-    #
+        self.widget.show()
+
     def on_fontpicker_font_set(self, picker):
         self.prefs.custom_font = picker.get_font_name()
 
@@ -202,6 +174,8 @@ class PreferencesDialog(gnomeglade.Component):
 
     def on_checkbutton_show_line_numbers_toggled(self, check):
         self.prefs.show_line_numbers = check.get_active()
+    def on_checkbutton_show_whitespace_toggled(self, check):
+        self.prefs.show_whitespace = check.get_active()
     def on_checkbutton_use_syntax_highlighting_toggled(self, check):
         self.prefs.use_syntax_highlighting = check.get_active()
 
@@ -242,6 +216,7 @@ class MeldPreferences(prefs.Preferences):
         "tab_size": prefs.Value(prefs.INT, 4),
         "spaces_instead_of_tabs": prefs.Value(prefs.BOOL, False),
         "show_line_numbers": prefs.Value(prefs.BOOL, 0),
+        "show_whitespace": prefs.Value(prefs.BOOL, False),
         "use_syntax_highlighting": prefs.Value(prefs.BOOL, 0),
         "edit_wrap_lines" : prefs.Value(prefs.INT, 0),
         "edit_command_type" : prefs.Value(prefs.STRING, "gnome"), #gnome, custom
@@ -250,19 +225,16 @@ class MeldPreferences(prefs.Preferences):
         "ignore_symlinks": prefs.Value(prefs.BOOL,0),
         "vc_console_visible": prefs.Value(prefs.BOOL, 0),
         "color_delete_bg" : prefs.Value(prefs.STRING, "DarkSeaGreen1"),
-        "color_delete_fg" : prefs.Value(prefs.STRING, "Red"),
         "color_replace_bg" : prefs.Value(prefs.STRING, "#ddeeff"),
-        "color_replace_fg" : prefs.Value(prefs.STRING, "Black"),
         "color_conflict_bg" : prefs.Value(prefs.STRING, "Pink"),
-        "color_conflict_fg" : prefs.Value(prefs.STRING, "Black"),
         "color_inline_bg" : prefs.Value(prefs.STRING, "LightSteelBlue2"),
         "color_inline_fg" : prefs.Value(prefs.STRING, "Red"),
-        "color_edited_bg" : prefs.Value(prefs.STRING, "gray90"),
-        "color_edited_fg" : prefs.Value(prefs.STRING, "Black"),
         "filters" : prefs.Value(prefs.STRING,
             #TRANSLATORS: translate this string ONLY to the first "\t", leave it and the following parts intact
             _("Backups\t1\t#*# .#* ~* *~ *.{orig,bak,swp}\n") + \
             #TRANSLATORS: translate this string ONLY to the first "\t", leave it and the following parts intact
+            _("OS-specific metadata\t0\t.DS_Store ._* .Spotlight-V100 .Trashes Thumbs.db Desktop.ini\n") + \
+            #TRANSLATORS: translate this string ONLY to the first "\t", leave it and the following parts intact
             _("Version Control\t1\t%s\n") % misc.shell_escape(' '.join(vc.get_plugins_metadata())) + \
             #TRANSLATORS: translate this string ONLY to the first "\t", leave it and the following parts intact
             _("Binaries\t1\t*.{pyc,a,obj,o,so,la,lib,dll}\n") + \
@@ -282,7 +254,11 @@ class MeldPreferences(prefs.Preferences):
             _("Script comment\t0\t#.*")),
         "ignore_blank_lines" : prefs.Value(prefs.BOOL, False),
         "toolbar_visible" : prefs.Value(prefs.BOOL, True),
-        "statusbar_visible" : prefs.Value(prefs.BOOL, True)
+        "statusbar_visible" : prefs.Value(prefs.BOOL, True),
+        "dir_status_filters": prefs.Value(prefs.LIST,
+                                          ['normal', 'modified', 'new']),
+        "vc_status_filters": prefs.Value(prefs.LIST,
+                                         ['flatten', 'modified']),
     }
 
     def __init__(self):
@@ -298,32 +274,41 @@ class MeldPreferences(prefs.Preferences):
 
     def get_toolbar_style(self):
         if not hasattr(self, "_gconf"):
-            return gtk.TOOLBAR_BOTH
-        style = self._gconf.get_string('/desktop/gnome/interface/toolbar_style') or "both"
-        style = {"both":gtk.TOOLBAR_BOTH, "text":gtk.TOOLBAR_TEXT,
-                 "icon":gtk.TOOLBAR_ICONS, "icons":gtk.TOOLBAR_ICONS,
-                 "both_horiz":gtk.TOOLBAR_BOTH_HORIZ,
-                 "both-horiz":gtk.TOOLBAR_BOTH_HORIZ
-                 }[style]
-        return style
-
-    def get_gnome_editor_command(self, files):
-        if not hasattr(self, "_gconf"):
-            return []
-        argv = []
-        editor = self._gconf.get_string('/desktop/gnome/applications/editor/exec') or "gedit"
-        if self._gconf.get_bool("/desktop/gnome/applications/editor/needs_term"):
-            texec = self._gconf.get_string("/desktop/gnome/applications/terminal/exec")
-            if texec:
-                argv.append(texec)
-                targ = self._gconf.get_string("/desktop/gnome/applications/terminal/exec_arg")
-                if targ:
-                    argv.append(targ)
-            argv.append( "%s %s" % (editor, " ".join( [f.replace(" ","\\ ") for f in files]) ) )
+            style = "both-horiz"
         else:
-            argv = [editor] + files
-        return argv
+            style = self._gconf.get_string(
+                      '/desktop/gnome/interface/toolbar_style') or "both-horiz"
+        toolbar_styles = {
+            "both": gtk.TOOLBAR_BOTH, "text": gtk.TOOLBAR_TEXT,
+            "icon": gtk.TOOLBAR_ICONS, "icons": gtk.TOOLBAR_ICONS,
+            "both_horiz": gtk.TOOLBAR_BOTH_HORIZ,
+            "both-horiz": gtk.TOOLBAR_BOTH_HORIZ
+        }
+        return toolbar_styles[style]
 
-    def get_custom_editor_command(self, files):
-        return self.edit_command_custom.split() + files
+    def get_editor_command(self, files, command_type=None):
+        if command_type is None:
+            command_type = self.edit_command_type
 
+        if command_type == "custom":
+            return self.edit_command_custom.split() + files
+        else:
+            if not hasattr(self, "_gconf"):
+                return []
+
+            editor_path = "/desktop/gnome/applications/editor/"
+            terminal_path = "/desktop/gnome/applications/terminal/"
+            editor = self._gconf.get_string(editor_path + "exec") or "gedit"
+            if self._gconf.get_bool(editor_path + "needs_term"):
+                argv = []
+                texec = self._gconf.get_string(terminal_path + "exec")
+                if texec:
+                    argv.append(texec)
+                    targ = self._gconf.get_string(terminal_path + "exec_arg")
+                    if targ:
+                        argv.append(targ)
+                escaped_files = [f.replace(" ", "\\ ") for f in files]
+                argv.append("%s %s" % (editor, " ".join(escaped_files)))
+                return argv
+            else:
+                return [editor] + files
