@@ -1,4 +1,5 @@
 ### Copyright (C) 2009 Vincent Legoll <vincent.legoll@gmail.com>
+### Copyright (C) 2010-2011 Kai Willadsen <kai.willadsen@gmail.com>
 
 ### This program is free software; you can redistribute it and/or modify
 ### it under the terms of the GNU General Public License as published by
@@ -19,12 +20,18 @@
 
 import os
 
+import gio
+import gtk
+
+
 class _srcviewer(object):
     # Module name to be imported for the sourceviewer class
     srcviewer_module = None
     # instance of the imported sourceviewer module
     gsv = None
 
+    spaces_flag = 0
+
     def __init__(self):
         if self.srcviewer_module is not None:
             self.gsv = __import__(self.srcviewer_module)
@@ -33,22 +40,25 @@ class _srcviewer(object):
         self.GtkTextView = None
         self.GtkTextBuffer = None
         self.overrides()
-        self.override = {}
-        self.override["GtkTextView"] = self.GtkTextView
-        self.override["GtkTextBuffer"] = self.GtkTextBuffer
 
     def version_check(self):
         raise NotImplementedError
+
     def overrides(self):
         raise NotImplementedError
+
     def GtkLanguageManager(self):
         raise NotImplementedError
-    def get_language_from_file(self, filename):
-        raise NotImplementedError
-    def set_highlight(self, buf, enabled):
+
+    def set_highlight_syntax(self, buf, enabled):
         raise NotImplementedError
+
     def set_tab_width(self, tab, tab_size):
         raise NotImplementedError
+
+    def get_language_from_file(self, filename):
+        raise NotImplementedError
+
     def get_language_from_mime_type(self, mimetype):
         raise NotImplementedError
 
@@ -57,52 +67,9 @@ class _srcviewer(object):
             self.glm = self.GtkLanguageManager()
         return self.glm
 
-    def set_highlighting_enabled(self, buf, gsl, enabled):
-        if enabled:
-            if gsl:
-                buf.set_language(gsl)
-            else:
-                enabled = False
-        self.set_highlight(buf, enabled)
+    def set_language(self, buf, lang):
+        raise NotImplementedError
 
-    def set_highlighting_enabled_from_mimetype(self, buf, mimetype, enabled):
-        self.set_highlighting_enabled(buf, self.get_language_from_mime_type(mimetype), enabled)
-
-    def set_highlighting_enabled_from_file(self, buf, fname, enabled):
-        self.set_highlighting_enabled(buf, self.get_language_from_file(os.path.abspath(fname)), enabled)
-
-class sourceview(_srcviewer):
-    srcviewer_module = "sourceview"
-
-    def version_check(self):
-        # ImportError exceptions are caught, so we
-        # won't use 'sourceview' without gnomevfs
-        import gnomevfs
-        self.gvfs = gnomevfs
-
-    def overrides(self):
-        self.GtkTextView = self.gsv.SourceView
-        self.GtkTextBuffer = self.gsv.SourceBuffer
-
-    def GtkLanguageManager(self):
-        return self.gsv.SourceLanguagesManager()
-
-    def set_tab_width(self, tab, tab_size):
-        return tab.set_tabs_width(tab_size)
-
-    def set_highlight(self, buf, enabled):
-        return buf.set_highlight(enabled)
-
-    def get_language_from_mime_type(self, mime_type):
-        return self.get_language_manager().get_language_from_mime_type(mime_type)
-
-    def get_language_from_file(self, filename):
-        mime_type = self.gvfs.get_mime_type(
-                self.gvfs.make_uri_from_input(os.path.abspath(filename)))
-        return self.get_language_from_mime_type(mime_type)
-
-class gtksourceview(sourceview):
-    srcviewer_module = "gtksourceview"
 
 class _gtksourceview2(_srcviewer):
     srcviewer_module = "gtksourceview2"
@@ -113,6 +80,7 @@ class _gtksourceview2(_srcviewer):
     def overrides(self):
         self.GtkTextView = self.gsv.View
         self.GtkTextBuffer = self.gsv.Buffer
+        self.spaces_flag = self.gsv.DRAW_SPACES_ALL
 
     def GtkLanguageManager(self):
         return self.gsv.LanguageManager()
@@ -120,34 +88,19 @@ class _gtksourceview2(_srcviewer):
     def set_tab_width(self, tab, tab_size):
         return tab.set_tab_width(tab_size)
 
-    def set_highlight(self, buf, enabled):
+    def set_highlight_syntax(self, buf, enabled):
         return buf.set_highlight_syntax(enabled)
 
     def get_language_from_file(self, filename):
-        raise NotImplementedError
+        return self.get_language_manager().guess_language(filename)
 
     def get_language_from_mime_type(self, mime_type):
-        for idl in self.get_language_manager().get_language_ids():
-            lang = self.get_language_manager().get_language(idl)
-            for mimetype in lang.get_mime_types():
-                if mime_type == mimetype:
-                    return lang
-        return None
+        content_type = gio.content_type_from_mime_type(mime_type)
+        return self.get_language_manager().guess_language(None, content_type)
 
-class gtksourceview22(_gtksourceview2):
+    def set_language(self, buf, lang):
+        buf.set_language(lang)
 
-    def version_check(self):
-        if self.gsv.pygtksourceview2_version[1] > 2:
-            raise ImportError
-
-    def get_language_from_file(self, filename):
-        from fnmatch import fnmatch
-        for idl in self.get_language_manager().get_language_ids():
-            lang = self.get_language_manager().get_language(idl)
-            for aglob in lang.get_globs():
-                if fnmatch(filename, aglob):
-                    return lang
-        return None
 
 class gtksourceview24(_gtksourceview2):
 
@@ -155,15 +108,52 @@ class gtksourceview24(_gtksourceview2):
         if self.gsv.pygtksourceview2_version[1] < 4:
             raise ImportError
 
-    def get_language_from_file(self, filename):
-        return self.get_language_manager().guess_language(filename)
+    def overrides(self):
+        _gtksourceview2.overrides(self)
+        viewClass = self.gsv.View
+
+        class SourceView(viewClass):
+
+            __gsignals__ = {
+                'key-press-event': 'override'
+            }
+
+            def do_key_press_event(self, event):
+                if event.keyval in (gtk.keysyms.KP_Up, gtk.keysyms.KP_Down,
+                                    gtk.keysyms.Up, gtk.keysyms.Down) and \
+                   (event.state & gtk.gdk.MOD1_MASK) != 0 and \
+                   (event.state & gtk.gdk.SHIFT_MASK) == 0:
+                    return True
+                return viewClass.do_key_press_event(self, event)
+
+        self.GtkTextView = SourceView
+
+
+class gtksourceview210(_gtksourceview2):
+
+    def version_check(self):
+        if self.gsv.pygtksourceview2_version[1] < 10:
+            raise ImportError
+
+    def overrides(self):
+        _gtksourceview2.overrides(self)
+        gtk.binding_entry_remove(self.GtkTextView, gtk.keysyms.Up,
+                                 gtk.gdk.MOD1_MASK)
+        gtk.binding_entry_remove(self.GtkTextView, gtk.keysyms.KP_Up,
+                                 gtk.gdk.MOD1_MASK)
+        gtk.binding_entry_remove(self.GtkTextView, gtk.keysyms.Down,
+                                 gtk.gdk.MOD1_MASK)
+        gtk.binding_entry_remove(self.GtkTextView, gtk.keysyms.KP_Down,
+                                 gtk.gdk.MOD1_MASK)
+
 
 class nullsourceview(_srcviewer):
     """Implement the sourceviewer API when no real one is available
     """
 
     get_language_from_file = lambda *args: None
-    set_highlight = lambda *args: None
+    set_highlight_syntax = lambda *args: None
+    set_language = lambda *args: None
     set_tab_width = lambda *args: None
     get_language_from_mime_type = lambda *args: None
 
@@ -175,6 +165,7 @@ class nullsourceview(_srcviewer):
             set_tab_width = lambda *args: None
             set_show_line_numbers = lambda *args: None
             set_insert_spaces_instead_of_tabs = lambda *args: None
+            set_draw_spaces = lambda *args: None
         gobject.type_register(NullTextView)
 
         self.GtkTextView = NullTextView
@@ -184,7 +175,7 @@ class nullsourceview(_srcviewer):
         pass
 
 def _get_srcviewer():
-    for srcv in (gtksourceview24, gtksourceview22, gtksourceview, sourceview):
+    for srcv in (gtksourceview210, gtksourceview24):
         try:
             return srcv()
         except ImportError:
@@ -192,3 +183,17 @@ def _get_srcviewer():
     return nullsourceview()
 
 srcviewer = _get_srcviewer()
+
+class MeldSourceView(srcviewer.GtkTextView):
+    __gtype_name__ = "MeldSourceView"
+
+    def get_y_for_line_num(self, line):
+        buf = self.get_buffer()
+        it = buf.get_iter_at_line(line)
+        y, h = self.get_line_yrange(it)
+        if line >= buf.get_line_count():
+            return y + h - 1
+        return y
+
+    def get_line_num_for_y(self, y):
+        return self.get_line_at_y(y)[0].get_line()
