@@ -1,4 +1,5 @@
 ### Copyright (C) 2002-2006 Stephen Kennedy <stevek@gnome.org>
+### Copyright (C) 2010-2012 Kai Willadsen <kai.willadsen@gmail.com>
 
 ### This program is free software; you can redistribute it and/or modify
 ### it under the terms of the GNU General Public License as published by
@@ -110,13 +111,20 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
                          "VcRevert": ("revert_command", ()),
                          }
 
+    state_actions = {
+        "flatten": ("VcFlatten", None),
+        "modified": ("VcShowModified", entry_modified),
+        "normal": ("VcShowNormal", entry_normal),
+        "unknown": ("VcShowNonVC", entry_nonvc),
+        "ignored": ("VcShowIgnored", entry_ignored),
+    }
+
     def __init__(self, prefs):
         melddoc.MeldDoc.__init__(self, prefs)
         gnomeglade.Component.__init__(self, paths.ui_dir("vcview.ui"), "vcview")
 
         actions = (
             ("VcCompare",       gtk.STOCK_DIALOG_INFO,      _("_Compare"),      None, _("Compare selected"), self.on_button_diff_clicked),
-            ("VcOpen",          gtk.STOCK_OPEN,             None,               None, _("Open selected"), self.on_button_open_clicked),
             ("VcCommit",        "vc-commit-24",             _("Co_mmit"),       None, _("Commit"), self.on_button_commit_clicked),
             ("VcUpdate",        "vc-update-24",             _("_Update"),       None, _("Update"), self.on_button_update_clicked),
             ("VcAdd",           "vc-add-24",                _("_Add"),          None, _("Add to VC"), self.on_button_add_clicked),
@@ -128,11 +136,11 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         )
 
         toggleactions = (
-            ("VcFlatten",     gtk.STOCK_GOTO_BOTTOM, _("_Flatten"),  None, _("Flatten directories"), self.on_button_flatten_toggled, True),
-            ("VcShowModified","filter-modified-24",  _("_Modified"), None, _("Show modified"), self.on_button_filter_toggled, True),
-            ("VcShowNormal",  "filter-normal-24",    _("_Normal"),   None, _("Show normal"), self.on_button_filter_toggled, False),
-            ("VcShowNonVC",   "filter-nonvc-24",     _("Non _VC"),   None, _("Show unversioned files"), self.on_button_filter_toggled, False),
-            ("VcShowIgnored", "filter-ignored-24",   _("Ignored"),   None, _("Show ignored files"), self.on_button_filter_toggled, False),
+            ("VcFlatten",     gtk.STOCK_GOTO_BOTTOM, _("_Flatten"),  None, _("Flatten directories"), self.on_button_flatten_toggled, False),
+            ("VcShowModified","filter-modified-24",  _("_Modified"), None, _("Show modified"), self.on_filter_state_toggled, False),
+            ("VcShowNormal",  "filter-normal-24",    _("_Normal"),   None, _("Show normal"), self.on_filter_state_toggled, False),
+            ("VcShowNonVC",   "filter-nonvc-24",     _("Non _VC"),   None, _("Show unversioned files"), self.on_filter_state_toggled, False),
+            ("VcShowIgnored", "filter-ignored-24",   _("Ignored"),   None, _("Show ignored files"), self.on_filter_state_toggled, False),
         )
 
         self.ui_file = paths.ui_dir("vcview-ui.xml")
@@ -154,6 +162,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         self.treeview.get_selection().set_mode(gtk.SELECTION_MULTIPLE)
         self.treeview.set_headers_visible(1)
         self.treeview.set_search_equal_func(self.treeview_search_cb)
+        self.prev_path, self.next_path = None, None
         column = gtk.TreeViewColumn( _("Name") )
         renicon = ui.emblemcellrenderer.EmblemCellRenderer()
         rentext = gtk.CellRendererText()
@@ -180,6 +189,14 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         addCol(_("Tag"), COL_TAG)
         addCol(_("Options"), COL_OPTIONS)
 
+
+        self.state_filters = []
+        for s in self.state_actions:
+            if s in self.prefs.vc_status_filters:
+                action_name = self.state_actions[s][0]
+                self.state_filters.append(s)
+                self.actiongroup.get_action(action_name).set_active(True)
+
         class ConsoleStream(object):
             def __init__(self, textview):
                 self.textview = textview
@@ -212,8 +229,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
 
     def on_container_switch_in_event(self, ui):
         melddoc.MeldDoc.on_container_switch_in_event(self, ui)
-        # FIXME: Add real sensitivity handling
-        self.emit("next-diff-changed", True, True)
+        self.scheduler.add_task(self.on_treeview_cursor_changed)
 
     def update_actions_sensitivity(self):
         """Disable actions that use not implemented VC plugin methods
@@ -294,6 +310,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         # need to scan the rest of the repository
         if os.path.isdir(self.vc.location):
             self.scheduler.add_task(self._search_recursively_iter(self.model.get_iter_root()).next)
+            self.scheduler.add_task(self.on_treeview_cursor_changed)
 
     def recompute_label(self):
         self.label_text = os.path.basename(self.location)
@@ -307,15 +324,11 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         rootname = self.model.value_path( self.model.get_iter(rootpath), 0 )
         prefixlen = 1 + len( self.model.value_path( self.model.get_iter_root(), 0 ) )
         todo = [ (rootpath, rootname) ]
-        filters = []
-        if self.actiongroup.get_action("VcShowModified").get_active():
-            filters.append( entry_modified )
-        if self.actiongroup.get_action("VcShowNormal").get_active():
-            filters.append( entry_normal )
-        if self.actiongroup.get_action("VcShowNonVC").get_active():
-            filters.append( entry_nonvc )
-        if self.actiongroup.get_action("VcShowIgnored").get_active():
-            filters.append( entry_ignored )
+
+        active_action = lambda a: self.actiongroup.get_action(a).get_active()
+        filters = [a[1] for a in self.state_actions.values() if \
+                   active_action(a[0]) and a[1]]
+
         def showable(entry):
             for f in filters:
                 if f(entry): return 1
@@ -415,17 +428,25 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
     def on_button_press_event(self, text, event):
         if event.button == 3:
             self.popup_menu.popup(None, None, None, event.button, event.time)
-            return len(self._get_selected_paths()) != 1
-        return 0
+            return True
+        return False
 
     def on_button_flatten_toggled(self, button):
-        self.treeview_column_location.set_visible(self.actiongroup.get_action("VcFlatten").get_active())
-        self.refresh()
-    def on_button_filter_toggled(self, button):
-        self.refresh()
+        action = self.actiongroup.get_action("VcFlatten")
+        self.treeview_column_location.set_visible(action.get_active())
+        self.on_filter_state_toggled(button)
 
-    def _get_selected_paths(self):
-        return self.treeview.get_selection().get_selected_rows()[1]
+    def on_filter_state_toggled(self, button):
+        active_action = lambda a: self.actiongroup.get_action(a).get_active()
+        active_filters = [a for a in self.state_actions if \
+                          active_action(self.state_actions[a][0])]
+
+        if set(active_filters) == set(self.state_filters):
+            return
+
+        self.state_filters = active_filters
+        self.prefs.vc_status_filters = active_filters
+        self.refresh()
 
     def _get_selected_files(self):
         sel = []
@@ -516,7 +537,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         if len(files):
             self.run_diff(files)
 
-    def on_button_open_clicked(self, obj):
+    def open_external(self):
         self._open_files(self._get_selected_files())
 
     def show_patch(self, prefix, patch, silent=False):
@@ -661,17 +682,25 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         item.show()
         menu.insert( item, 1 )
 
-    def next_diff(self, direction):
-        start_iter = self.model.get_iter((self._get_selected_paths() or [(0,)])[-1])
+    def on_treeview_cursor_changed(self, *args):
+        cursor_path, cursor_col = self.treeview.get_cursor()
+        if not cursor_path:
+            self.emit("next-diff-changed", False, False)
+        else:
+            # TODO: Only recalculate when needed; see DirDiff
+            prev_path, next_path = self.model._find_next_prev_diff(cursor_path)
+            self.prev_path, self.next_path = prev_path, next_path
+            have_next_diffs = (prev_path is not None, next_path is not None)
+            self.emit("next-diff-changed", *have_next_diffs)
 
-        search = {gtk.gdk.SCROLL_UP : self.model.inorder_search_up}.get(direction, self.model.inorder_search_down)
-        for it in search( start_iter ):
-            state = self.model.get_state(it, 0)
-            if state not in (tree.STATE_NORMAL, tree.STATE_EMPTY):
-                curpath = self.model.get_path(it)
-                self.treeview.expand_to_path(curpath)
-                self.treeview.set_cursor(curpath)
-                return
+    def next_diff(self, direction):
+        if direction == gtk.gdk.SCROLL_UP:
+            path = self.prev_path
+        else:
+            path = self.next_path
+        if path:
+            self.treeview.expand_to_path(path)
+            self.treeview.set_cursor(path)
 
     def on_reload_activate(self, *extra):
         self.on_fileentry_activate(self.fileentry)
